package FeldmanVerifiableSecretSharing;

import ShamirSecretSharing.Polynomial;
import ShamirSecretSharing.SecretSharing;

import java.util.Arrays;

import org.factcenter.qilin.primitives.Group;
import org.factcenter.qilin.primitives.concrete.Zpstar;

import java.math.BigInteger;
import java.util.Random;

/**
 * Created by Tzlil on 1/27/2016.
 *
 * an implementation of Feldman's verifiable secret sharing scheme.
 *
 * allows trusted dealer to share a key x among n parties.
 *
 */
public class VerifiableSecretSharing extends SecretSharing {
    protected final Group<BigInteger> group;
    protected final BigInteger g; // public generator of group
    protected final BigInteger[] commitmentsArray;
    /**
     * @param group
     * @param q a large prime dividing group order.
     * @param g a generator of cyclic group of order q.
     *          the generated group is a subgroup of the given group.
     *          it must be chosen such that computing discrete logarithms is hard in this group.
     */
    public VerifiableSecretSharing(int t, int n, BigInteger x, Random random, BigInteger q, BigInteger g
            , Group<BigInteger> group) {
        super(t, n, x, random,q);
        this.g = g;
        this.group = group;
        assert (this.group.contains(g));
        this.commitmentsArray = generateCommitments();
    }

    /**
     * @return commitments[i] = g ^ polynomial.coefficients[i]
     */
    private BigInteger[] generateCommitments() {

        Polynomial polynomial = getPolynomial();
        BigInteger[] coefficients = polynomial.getCoefficients();
        BigInteger[] commitments = new BigInteger[coefficients.length];
        for (int i = 0 ; i < commitments.length;i++){
            commitments[i] = group.multiply(g,coefficients[i]);
        }
        return commitments;
    }

    /**
     * @param j share holder id
     * @param commitments
     * @param group
     *
     * @return product of commitmentsArray[j] ^ (i ^ j) == g ^ polynomial(i)
     */
    public static BigInteger verify(int j,BigInteger[] commitments,Group<BigInteger> group) {
        BigInteger v = group.zero();
        BigInteger power = BigInteger.ONE;
        BigInteger J = BigInteger.valueOf(j);
        for (int k = 0 ; k < commitments.length ; k ++){
            v = group.add(v,group.multiply(commitments[k],power));
            power = power.multiply(J);
        }
        return v;
    }


    /**
     * getter
     * @return generator of group
     */
    public BigInteger getGenerator() {
        return g;
    }

    /**
     * getter
     * @return group
     */
    public Group<BigInteger> getGroup(){
        return group;
    }

    /**
     * getter
     * @return copy of commitmentsArray
     */
    public BigInteger[] getCommitmentsArray() {
        return Arrays.copyOf(commitmentsArray, commitmentsArray.length);
    }

}
