package JointFeldmanProtocol;

import Communication.User;
import ShamirSecretSharing.Polynomial;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import com.google.protobuf.ByteString;
import meerkat.protobuf.DKGMessages.*;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 2/5/2016.
 *
 * an implementation of a version of Pedersen's distributed key generation protocol
 */
public class DistributedKeyGeneration extends VerifiableSecretSharing{

    protected final int id;
    protected Polynomial.Point[] shares;

    public DistributedKeyGeneration(int t, int n, BigInteger zi, Random random, BigInteger q, BigInteger g
            , Group<BigInteger> group, int id)  {
        super(t, n, zi, random, q, g,group);
        this.id = id;
        this.shares = null;
    }

    /**
     * stage1.1 according to the protocol
     *  Pi broadcasts Aik for k = 0,...,t.
     */
    public void broadcastCommitments(User user){
        broadcastCommitments(user,commitmentsArray);
    }

    public void broadcastCommitments(User user, BigInteger[] commitments){
        CommitmentMessage commitmentMessage;
        for (int k = 0; k <= t ; k++){
            commitmentMessage = CommitmentMessage.newBuilder()
                    .setCommitment(ByteString.copyFrom(commitments[k].toByteArray()))
                    .setK(k)
                    .build();
            user.broadcast(Mail.Type.COMMITMENT, commitmentMessage);
        }
    }

    public void sendSecret(User user, int j){
        SecretMessage.Point secret = getShare(j).asMessage();
        user.send(j, Mail.Type.DOUBLE,SecretMessage.newBuilder().setSecret(secret).build());
    }

    /**
     * stage1.2 according to the protocol
     *  Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    public void sendSecrets(User user){
        for (int j = 1; j <= n ; j++){
            if(j != id){
                sendSecret(user,j);
            }
        }
    }

    public boolean isValidSecret(int j,BigInteger[] commitments,int i){
        Polynomial.Point secret = shares[j - 1];
        return isValidSecret(secret,commitments,i);
    }

    public boolean isValidSecret(Polynomial.Point secret, BigInteger[] commitments, int i){
        BigInteger v = verify(i,commitments,group);
        return group.multiply(g,secret.y).equals(v);
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received (using isValidSecret)
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     */
    public void broadcastComplains(User user, BigInteger[][]commitmentsTable){
        ComplaintMessage complaint;
        for (int j = 1; j <= n ; j++ ){
            if(j != id) {
                if (!isValidSecret(j,commitmentsTable[j - 1],id)) {
                    //message = new Message(Type.Complaint, j)
                    complaint = ComplaintMessage.newBuilder()
                            .setId(j)
                            .build();
                    user.broadcast(Mail.Type.COMPLAINT, complaint);
                }
            }
        }
    }


    public void broadcastComplaintAnswer(User user, int j){
        user.broadcast(Mail.Type.SECRET, SecretMessage.newBuilder()
                .setSecret(getShare(j).asMessage())
                .build());
    }

    /**
     * stage3.1 according to the protocol
     *   if more than t players complain against a player Pi he is disqualified.
     */
    public void answerAllComplainingPlayers(User user, DistributedKeyGenerationUserImpl.ComplainState[] complains){
        for (int j = 1; j <= n ; j++) {
            switch (complains[j - 1]) {
                case Waiting:
                    broadcastComplaintAnswer(user,j);
                    break;
                default:
                    break;
            }
        }
    }

    /**
     * stage3.2 according to the protocol
     *  if any of the revealed shares fails the verification test, player Pi is disqualified.
     *  set QUAL to be the set of non-disqualified players.
     */
    public Set<Integer> calcQUAL(DistributedKeyGenerationUserImpl.ComplainState[][] complains){
        Set<Integer> QUAL = new HashSet<Integer>();
        boolean nonDisqualified;
        int counter;
        for (int i = 1; i <= complains.length; i++){
            nonDisqualified = true;
            counter = 0;
            for (int j = 1; j <= complains[i - 1].length; j++){
                switch (complains[i - 1][j - 1]) {
                    case Non:
                        break;
                    case NonDisqualified:
                        counter++;
                    default:
                        nonDisqualified = false;
                }
                if(!nonDisqualified)
                    break;
            }
            if(nonDisqualified && counter <= t){
                QUAL.add(i);
            }
        }
        return QUAL;
    }

    /**
     * stage4.1 according to the protocol
     *   public value y is computed as y = multiplication of yi mod p for i in QUAL
     */
    public BigInteger calcY(BigInteger[] ys,Set<Integer> QUAL){
        BigInteger y = group.zero();
        for (int i : QUAL) {
            y = group.add(y , ys[i - 1]);
        }
        return y;
    }

    /**
     * stage4.2 according to the protocol
     *  public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     */
    public BigInteger[] calcCommitments(BigInteger[][] commitmentsTable,Set<Integer> QUAL){
        BigInteger[] commitments = new BigInteger[t + 1];
        Arrays.fill(commitments,group.zero());
        for (int i : QUAL) {
            for (int k = 0; k <= t; k++){
                commitments[k] = group.add(commitments[k],commitmentsTable[i - 1][k]);
            }
        }
        return commitments;
    }

    /**
     * stage4.3 according to the protocol
     *  Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    public Polynomial.Point calcShare(Polynomial.Point[] shares,Set<Integer> QUAL){
        BigInteger xj = BigInteger.ZERO;
        for (int i : QUAL) {
            xj = xj.add(shares[i - 1].y);
        }
        return new Polynomial.Point(BigInteger.valueOf(id) , xj.mod(q));
    }

    public int getId() {
        return id;
    }

    public void setShares(Polynomial.Point[] shares){
        this.shares = shares;
    }

}
