package JointFeldmanProtocol;

import Communication.Network;
import Communication.User;
import ShamirSecretSharing.Polynomial;
import UserInterface.DistributedKeyGenerationUser;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.Set;

/**
 * Created by Tzlil on 2/21/2016.
 */
public class DistributedKeyGenerationUserImpl implements DistributedKeyGenerationUser {
    protected enum ComplainState{
        Non, Waiting,Disqualified,NonDisqualified
    }

    protected final DistributedKeyGeneration dkg;

    protected final BigInteger g;
    protected final Group<BigInteger> group;
    protected final int n;
    protected final int t;
    protected final int id;

    protected MessageHandler messageHandler;
    protected final Polynomial.Point[] shares;
    protected final BigInteger[][] commitmentsTable;
    protected final boolean[] doneFlags;
    protected final User user;
    protected final ComplainState[][] complaintsTable;

    protected Set<Integer> QUAL; // set of all non-disqualified parties
    protected BigInteger[] commitments; // public verification values
    protected Polynomial.Point share; // final share of the secrete
    protected BigInteger y; // final public value

    public DistributedKeyGenerationUserImpl(DistributedKeyGeneration dkg, Network network) {
        this.dkg = dkg;

        this.g = dkg.getGenerator();
        this.group = dkg.getGroup();
        this.n = dkg.getN();
        this.t = dkg.getT();
        this.id = dkg.getId();

        this.messageHandler = new MessageHandler();
        this.shares = new Polynomial.Point[n];
        this.shares[id - 1] = dkg.getShare(id);
        this.commitmentsTable = new BigInteger[n][t + 1];
        this.doneFlags = new boolean[n];
        this.user = network.connect(messageHandler);
        this.complaintsTable = new ComplainState[n][n];
        for (int i = 0; i < n; i++){
            Arrays.fill(complaintsTable[i],ComplainState.Non);
        }

        this.QUAL = null;
        this.commitments = null;
        this.share = null;
        this.y = null;
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Aik for k = 0,...,t.
     *  2. Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    protected void stage1() {
        dkg.broadcastCommitments(user);
        dkg.sendSecrets(user);
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    protected void stage2(){
        dkg.setShares(shares);
        dkg.broadcastComplains(user,commitmentsTable);
        //broadcast done message after all complaints
        DKGMessages.DoneMessage doneMessage = DKGMessages.DoneMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);
    }

    /**
     * stage3 according to the protocol
     *  1. if more than t players complain against a player Pi he is disqualified.
     *     otherwise Pi broadcasts the share Sij for each complaining player Pj.
     *  2. if any of the revealed shares fails the verification test, player Pi is disqualified.
     *     set QUAL to be the set of non-disqualified players.
     */
    protected void stage3(){

        dkg.answerAllComplainingPlayers(user,complaintsTable[id - 1]);

        // wait until there is no complaint waiting for answer
        for (int i = 0; i < complaintsTable.length; i++){
            for (int j = 0; j < complaintsTable[i].length; j++){
                while (complaintsTable[i][j].equals(ComplainState.Waiting)){
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
        this.QUAL = dkg.calcQUAL(complaintsTable);
    }

    /**
     * stage4 according to the protocol
     *  1. public value y is computed as y = multiplication of yi mod p for i in QUAL
     *  2. public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     *  3. Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    protected void stage4(){
        BigInteger[] ys = new BigInteger[n];
        for (int i = 0; i < n; i++){
            ys[i] = commitmentsTable[i][0];
        }
        this.y = dkg.calcY(ys,QUAL);
        this.commitments = dkg.calcCommitments(commitmentsTable,QUAL);
        this.share = dkg.calcShare(shares,QUAL);
    }

    @Override
    public void run() {
        user.getReceiverThread().start();
        stage1();
        while (messageHandler.secretsCounter != n - 1 || messageHandler.commitmentsCounter != n * (t + 1)){
            try {
                Thread.sleep(300);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
        stage2();
        while (messageHandler.doneCounter != n){
            try {
                Thread.sleep(300);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
        stage3();
        stage4();
        user.getReceiverThread().interrupt();
    }

    @Override
    public BigInteger[] getCommitments() {
        return Arrays.copyOf(commitments, commitments.length);
    }

    @Override
    public BigInteger getGenerator() {
        return g;
    }

    @Override
    public Group<BigInteger> getGroup() {
        return group;
    }

    @Override
    public Polynomial.Point getShare() {
        return share;
    }

    @Override
    public int getID() {
        return id;
    }

    @Override
    public int getN() {
        return n;
    }

    @Override
    public int getT() {
        return t;
    }

    @Override
    public BigInteger getPublicValue() {
        return y;
    }

    @Override
    public Set<Integer> getQUAL() {
        return QUAL;
    }

    protected class MessageHandler implements Communication.MessageHandler{

        public int doneCounter;
        public int commitmentsCounter;
        public int secretsCounter;

        public MessageHandler() {
            this.doneCounter = 0;
            this.secretsCounter = 0;
            this.commitmentsCounter = 0;
        }

        @Override
        public void handelComplaintMessage(int sender, boolean isBroadcast, DKGMessages.ComplaintMessage complaintMessage) {
            if(isBroadcast) {
                int i = sender - 1;
                int j = complaintMessage.getId() - 1;
                switch (complaintsTable[i][j]) {
                    case Non:
                        complaintsTable[i][j] = ComplainState.Waiting;
                        break;
                    default:
                        break;
                }
            }
        }

        @Override
        public void handelDoneMessage(int sender, boolean isBroadcast, DKGMessages.DoneMessage doneMessage) {
            if(isBroadcast && !doneFlags[sender - 1]) {
                doneFlags[sender - 1] = true;
                doneCounter++;
            }
        }

        @Override
        public void handelCommitmentMessage(int sender, boolean isBroadcast, DKGMessages.CommitmentMessage commitmentMessage) {
            if(isBroadcast){
                int i = sender - 1;
                int k = commitmentMessage.getK();
                if(commitmentsTable[i][k] == null){
                    commitmentsTable[i][k] = extractCommitment(commitmentMessage);
                    commitmentsCounter++;
                }
            }
        }

        @Override
        public void handelSecretMessage(int sender, boolean isBroadcast, DKGMessages.SecretMessage secretMessage) {
            Polynomial.Point secret = extractSecret(secretMessage);
            if(!isBroadcast){
                if(shares[sender - 1] == null) {
                    shares[sender - 1] = secret;
                    secretsCounter++;
                }
            }else {
                int i = sender;
                int j = secret.x.intValue();
                switch (complaintsTable[i - 1][j - 1]){
                    case Waiting:
                        if(dkg.isValidSecret(secret,commitmentsTable[i - 1],j)){
                            complaintsTable[i - 1][j - 1] = ComplainState.NonDisqualified;
                        }else{
                            complaintsTable[i - 1][j - 1] = ComplainState.Disqualified;
                        }
                        break;
                    default:
                        break;
                }
            }
        }

        @Override
        public void handelDoubleSecretMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {

        }

        public Polynomial.Point extractSecret(DKGMessages.SecretMessage secretMessage){
            return new Polynomial.Point(secretMessage.getSecret());
        }

        public BigInteger extractCommitment(DKGMessages.CommitmentMessage commitmentMessage){
            return new BigInteger(commitmentMessage.getCommitment().toByteArray());
        }
    }
}
