package SecureDistributedKeyGeneration;

import Communication.Network;
import Communication.User;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import JointFeldmanProtocol.DistributedKeyGeneration;
import ShamirSecretSharing.Polynomial;
import com.google.protobuf.ByteString;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;
import org.factcenter.qilin.primitives.concrete.Zpstar;

import java.math.BigInteger;
import java.util.Random;

/**
 * Created by Tzlil on 2/17/2016.
 */
public class SecureDistributedKeyGeneration extends DistributedKeyGeneration {

    private VerifiableSecretSharing verifiableSecretSharing;
    private final BigInteger h;
    private Polynomial.Point[] sharesT;

    public SecureDistributedKeyGeneration(int t, int n, BigInteger zi, Random random, BigInteger q, BigInteger g
            , BigInteger h, Group<BigInteger> group, int id) {
        super(t, n, zi, random, q, g, group, id);
        this.h = h;
        BigInteger r = new BigInteger(q.bitLength(),random).mod(q);
        this.verifiableSecretSharing = new VerifiableSecretSharing(t,n,r,random,q,h,group);
    }

    @Override
    public void sendSecret(User user,int j) {
        DKGMessages.SecretMessage.Point secret = getShare(j).asMessage();;
        DKGMessages.SecretMessage.Point secretT = verifiableSecretSharing.getShare(j).asMessage();;
        DKGMessages.DoubleSecretMessage doubleSecretMessage = DKGMessages.DoubleSecretMessage.newBuilder()
                .setS1(DKGMessages.SecretMessage.newBuilder().setSecret(secret).build())
                .setS2(DKGMessages.SecretMessage.newBuilder().setSecret(secretT).build())
                .build();
        user.send(j, DKGMessages.Mail.Type.DOUBLE, doubleSecretMessage);
    }

    @Override
    public boolean isValidSecret(int j, BigInteger[] commitments, int i){
        Polynomial.Point secret = shares[j - 1];
        Polynomial.Point secretT = sharesT[j - 1];
        return isValidSecret(secret,secretT,commitments,i);
    }

    public boolean isValidSecret(Polynomial.Point secret,Polynomial.Point secretT, BigInteger[] verificationValues, int i){
        BigInteger v = verify(i,verificationValues,group);
        BigInteger exp = group.add(group.multiply(g, secret.y),group.multiply(h, secretT.y));
        return exp.equals(v);
    }

    public boolean isValidComplaint(Polynomial.Point secret,Polynomial.Point secretT, BigInteger[] commitments
            ,BigInteger[] verificationValues, int i){
        return isValidSecret(secret,secretT,verificationValues,i) && !isValidSecret(secret,commitments,i);
    }

    /**
     * stage4.3 according to the protocol
     *  if check fails for index i, Pj
     */
    public void broadcastComplaints(User user, BigInteger[][] commitmentsTable, boolean stage4){
        if(!stage4){
            broadcastComplains(user,commitmentsTable);
        }else{
            for (int j = 1; j <= n ; j++ ){
                if(j != id) {
                    if (!isValidSecret(shares[j - 1],commitmentsTable[j - 1],id)) {
                        broadcastDoubleSecret(user,shares[j - 1],sharesT[j - 1]);
                    }
                }
            }
        }
    }


    public void broadcastVerificationValues(User user){
        BigInteger[] verificationValues = new BigInteger[t + 1];
        BigInteger[] hBaseCommitments = verifiableSecretSharing.getCommitmentsArray();
        for (int k = 0 ; k < verificationValues.length ; k++){
            verificationValues[k] = group.add(commitmentsArray[k],hBaseCommitments[k]);
        }
        broadcastCommitments(user,verificationValues);
    }

    private void broadcastDoubleSecret(User user ,Polynomial.Point secret, Polynomial.Point secretT){
        DKGMessages.SecretMessage.Point secretMessage = secret.asMessage();
        DKGMessages.SecretMessage.Point secretTMessage = secretT.asMessage();
        DKGMessages.DoubleSecretMessage doubleSecretMessage = DKGMessages.DoubleSecretMessage.newBuilder()
                .setS1(DKGMessages.SecretMessage.newBuilder().setSecret(secretMessage).build())
                .setS1(DKGMessages.SecretMessage.newBuilder().setSecret(secretTMessage).build())
                .build();
        user.broadcast(DKGMessages.Mail.Type.DOUBLE, doubleSecretMessage);
    }

    @Override
    public void broadcastComplaintAnswer(User user, int j) {
        broadcastDoubleSecret(user,getShare(j),verifiableSecretSharing.getShare(j));
    }

    public void setSharesT(Polynomial.Point[] sharesT) {
        this.sharesT = sharesT;
    }
}
